##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Steam turbine stage model. This is a standard isentropic turbine. Under off-design
conditions the base efficiency and pressure ratio do not change much for the stages
between the inlet and outlet. This model is based on:

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136.
"""
__Author__ = "John Eslick"

import logging

_log = logging.getLogger(__name__)

from pyomo.common.config import In
from pyomo.environ import Var, Expression, SolverFactory, value
from pyomo.opt import TerminationCondition

from idaes.core import declare_process_block_class
from idaes.unit_models.pressure_changer import (
    PressureChangerData,
    ThermodynamicAssumption,
)
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.util.model_statistics import degrees_of_freedom


@declare_process_block_class("TurbineStage", doc="Basic steam turbine model")
class TurbineStageData(PressureChangerData):
    # Same settings as the default pressure changer, but force to expander with
    # isentropic efficiency
    CONFIG = PressureChangerData.CONFIG()
    CONFIG.compressor = False
    CONFIG.get("compressor")._default = False
    CONFIG.get("compressor")._domain = In([False])
    CONFIG.thermodynamic_assumption = ThermodynamicAssumption.isentropic
    CONFIG.get("thermodynamic_assumption")._default = ThermodynamicAssumption.isentropic
    CONFIG.get("thermodynamic_assumption")._domain = In(
        [ThermodynamicAssumption.isentropic]
    )

    def build(self):
        super(TurbineStageData, self).build()

        self.efficiency_mech = Var(initialize=0.98, doc="Turbine mechanical efficiency")
        self.efficiency_mech.fix()
        self.ratioP[:] = 0.8  # make sure these have a number value
        self.deltaP[:] = 0  #   to avoid an error later in initialize
        time_set = self.flowsheet().config.time

        self.shaft_speed = Var(time_set, doc="Shaft speed [1/s]", initialize=60.0)

        @self.Expression(time_set, doc="Specific speed [dimensionless]")
        def specific_speed(b, t):
            s = b.shaft_speed[t]  # 1/s
            v = b.control_volume.properties_out[t].flow_vol  # m3/s
            his_rate = b.work_isentropic[t]  # J/s
            m = b.control_volume.properties_out[t].flow_mass  # kg/s
            return s * v ** 0.5 * (his_rate / m) ** (-0.75)  # dimensionless

        @self.Expression(time_set, doc="Thermodynamic power [J/s]")
        def power_thermo(b, t):
            return b.control_volume.work[t]

        @self.Expression(self.flowsheet().config.time, doc="Shaft power [J/s]")
        def power_shaft(b, t):
            return b.power_thermo[t] * b.efficiency_mech

    def _get_performance_contents(self, time_point=0):
        pc = super()._get_performance_contents(time_point=time_point)
        pc["vars"]["Mechanical Efficiency"] = self.efficiency_mech

        return pc

    def initialize(
        self,
        state_args={},
        outlvl=6,
        solver="ipopt",
        optarg={"tol": 1e-6, "max_iter": 30},
    ):
        """
        Initialize the turbine stage model.  This deactivates the
        specialized constraints, then does the isentropic turbine initialization,
        then reactivates the constraints and solves.

        Args:
            state_args (dict): Initial state for property initialization
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        # sp is what to save to make sure state after init is same as the start
        #   saves value, fixed, and active state, doesn't load originally free
        #   values, this makes sure original problem spec is same but initializes
        #   the values of free vars
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # fix inlet and free outlet
        for t in self.flowsheet().config.time:
            for k, v in self.inlet.vars.items():
                v[t].fix()
            for k, v in self.outlet.vars.items():
                v[t].unfix()
            # If there isn't a good guess for efficiency or outlet pressure
            # provide something reasonable.
            eff = self.efficiency_isentropic[t]
            eff.fix(eff.value if value(eff) > 0.3 and value(eff) < 1.0 else 0.8)
            # for outlet pressure try outlet pressure, pressure ratio, delta P,
            # then if none of those look reasonable use a pressure ratio of 0.8
            # to calculate outlet pressure
            Pout = self.outlet.pressure[t]
            Pin = self.inlet.pressure[t]
            prdp = value((self.deltaP[t] - Pin) / Pin)
            if self.deltaP[t].fixed:
                Pout.value = value(Pin - Pout)
            if self.ratioP[t].fixed:
                Pout.value = value(self.ratioP[t] * Pin)
            if value(Pout / Pin) > 0.99 or value(Pout / Pin) < 0.1:
                if value(self.ratioP[t]) < 0.99 and value(self.ratioP[t]) > 0.1:
                    Pout.fix(value(Pin * self.ratioP[t]))
                elif prdp < 0.99 and prdp > 0.1:
                    Pout.fix(value(prdp * Pin))
                else:
                    Pout.fix(value(Pin * 0.8))
            else:
                Pout.fix()
            self.deltaP[t] = value(Pout - Pin)
            self.ratioP[t] = value(Pout / Pin)

        self.deltaP[:].unfix()
        self.ratioP[:].unfix()

        for t in self.flowsheet().config.time:
            self.properties_isentropic[t].pressure.value = value(
                self.outlet.pressure[t]
            )
            self.properties_isentropic[t].flow_mol.value = value(self.inlet.flow_mol[t])
            self.properties_isentropic[t].enth_mol.value = value(
                self.inlet.enth_mol[t] * 0.95
            )
            self.outlet.flow_mol[t].value = value(self.inlet.flow_mol[t])
            self.outlet.enth_mol[t].value = value(self.inlet.enth_mol[t] * 0.95)

        # Make sure the initialization problem has no degrees of freedom
        # This shouldn't happen here unless there is a bug in this
        dof = degrees_of_freedom(self)
        try:
            assert dof == 0
        except:
            _log.exception("degrees_of_freedom = {}".format(dof))
            raise

        # one bad thing about reusing this is that the log messages aren't
        # really compatible with being nested inside another initialization
        super(TurbineStageData, self).initialize(
            state_args=state_args, outlvl=outlvl, solver=solver, optarg=optarg
        )

        # reload original spec
        from_json(self, sd=istate, wts=sp)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Basic heater/cooler models
"""

__author__ = "John Eslick"

# Import Pyomo libraries
from pyomo.environ import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        MaterialBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.logger import getIdaesLogger

_log = getIdaesLogger(__name__)


def _make_heater_control_volume(o, name, config,
                                dynamic=None, has_holdup=None):
    """
    This is seperated from the main heater class so it can be reused to create
    control volumes for different types of heat exchange models.
    """
    if dynamic is None:
        dynamic = config.dynamic
    if has_holdup is None:
        has_holdup = config.has_holdup
    # we have to attach this control volume to the model for the rest of
    # the steps to work
    o.add_component(name, ControlVolume0DBlock(default={
        "dynamic": dynamic,
        "has_holdup": has_holdup,
        "property_package": config.property_package,
        "property_package_args": config.property_package_args}))
    control_volume = getattr(o, name)
    # Add inlet and outlet state blocks to control volume
    if has_holdup:
        control_volume.add_geometry()
    control_volume.add_state_blocks(
        has_phase_equilibrium=config.has_phase_equilibrium)
    # Add material balance
    control_volume.add_material_balances(
        balance_type=config.material_balance_type,
        has_phase_equilibrium=config.has_phase_equilibrium)
    # add energy balance
    control_volume.add_energy_balances(
        balance_type=config.energy_balance_type,
        has_heat_transfer=True)
    # add momentum balance
    control_volume.add_momentum_balances(
        balance_type=config.momentum_balance_type,
        has_pressure_change=config.has_pressure_change)
    return control_volume


def _make_heater_config_block(config):
    """
    Declare configuration options for HeaterData block.
    """
    config.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    config.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    config.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    config.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Phase equilibrium construction flag",
        doc="""Indicates whether terms for phase equilibrium should be
constructed, **default** = False.
**Valid values:** {
**True** - include phase equilibrium terms
**False** - exclude phase equilibrium terms.}"""))
    config.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    config.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    config.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))


@declare_process_block_class("Heater", doc="Simple 0D heater/cooler model.")
class HeaterData(UnitModelBlockData):
    """
    Simple 0D heater unit.
    Unit model to add or remove heat from a material.
    """
    CONFIG = UnitModelBlockData.CONFIG()
    _make_heater_config_block(CONFIG)

    def build(self):
        """Building model

        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(HeaterData, self).build()
        # Add Control Volume
        _make_heater_control_volume(self, "control_volume", self.config)
        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()
        # Add a convienient reference to heat duty.
        self.heat_duty = Reference(self.control_volume.heat)
        if (self.config.has_pressure_change is True and
            self.config.momentum_balance_type != MomentumBalanceType.none):

            self.deltaP = Reference(self.control_volume.deltaP)

    def _get_performance_contents(self, time_point=0):
        return {"vars":{"Heat Duty":self.heat_duty[time_point]}}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for Pressure Changer unit model.

Author: Andrew Lee, Emmanuel Ogbe
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           Var)

from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)

from idaes.unit_models.pressure_changer import (PressureChanger,
                                                ThermodynamicAssumption)

from idaes.property_models.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.property_models import iapws95
from idaes.property_models.examples.saponification_thermo import \
    SaponificationParameterBlock

from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)
from idaes.core.util.exceptions import BalanceTypeNotSupportedError


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_ThermodynamicAssumption():
    assert len(ThermodynamicAssumption) == 4


class TestPressureChanger(object):
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties})

        # Check unit config arguments
        assert len(m.fs.unit.config) == 10

        assert m.fs.unit.config.material_balance_type == \
            MaterialBalanceType.useDefault
        assert m.fs.unit.config.energy_balance_type == \
            EnergyBalanceType.useDefault
        assert m.fs.unit.config.momentum_balance_type == \
            MomentumBalanceType.pressureTotal
        assert not m.fs.unit.config.has_phase_equilibrium
        assert m.fs.unit.config.compressor
        assert m.fs.unit.config.thermodynamic_assumption == \
            ThermodynamicAssumption.isothermal
        assert m.fs.unit.config.property_package is m.fs.properties

    def test_dynamic_build(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": True})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties})

        assert hasattr(m.fs.unit, "volume")

    def test_pump(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.pump})

        assert isinstance(m.fs.unit.fluid_work_calculation, Constraint)

    def test_adiabatic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.adiabatic})

        assert isinstance(m.fs.unit.adiabatic, Constraint)

    def test_isentropic_comp_phase_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.componentPhase})

        assert isinstance(m.fs.unit.isentropic_material, Constraint)
        assert len(m.fs.unit.isentropic_material) == 4

    def test_isentropic_comp_total_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.componentTotal})

        assert isinstance(m.fs.unit.isentropic_material, Constraint)
        assert len(m.fs.unit.isentropic_material) == 2

    def test_isentropic_total_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(BalanceTypeNotSupportedError):
            m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.total})

    def test_isentropic_total_element_balances(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(BalanceTypeNotSupportedError):
            m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.elementTotal})

    def test_isentropic_material_balances_none(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(BalanceTypeNotSupportedError):
            m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "material_balance_type": MaterialBalanceType.none})


# -----------------------------------------------------------------------------
class TestBTX_isothermal(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase": 'Liq'})

        m.fs.unit = PressureChanger(default={
            "property_package": m.fs.properties,
            "thermodynamic_assumption": ThermodynamicAssumption.isothermal})

        return m

    @pytest.mark.build
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "inlet")
        assert len(btx.fs.unit.inlet.vars) == 4
        assert hasattr(btx.fs.unit.inlet, "flow_mol")
        assert hasattr(btx.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet, "temperature")
        assert hasattr(btx.fs.unit.inlet, "pressure")

        assert hasattr(btx.fs.unit, "outlet")
        assert len(btx.fs.unit.outlet.vars) == 4
        assert hasattr(btx.fs.unit.outlet, "flow_mol")
        assert hasattr(btx.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet, "temperature")
        assert hasattr(btx.fs.unit.outlet, "pressure")

        assert hasattr(btx.fs.unit, "work_mechanical")
        assert hasattr(btx.fs.unit, "deltaP")
        assert isinstance(btx.fs.unit.ratioP, Var)
        assert isinstance(btx.fs.unit.ratioP_calculation, Constraint)

        assert number_variables(btx) == 25
        assert number_total_constraints(btx) == 19
        assert number_unused_variables(btx) == 0

    def test_dof(self, btx):
        btx.fs.unit.inlet.flow_mol[0].fix(5)  # mol/s
        btx.fs.unit.inlet.temperature[0].fix(365)  # K
        btx.fs.unit.inlet.pressure[0].fix(101325)  # Pa
        btx.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.deltaP.fix(50000)

        assert degrees_of_freedom(btx) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, btx):
        orig_fixed_vars = fixed_variables_set(btx)
        orig_act_consts = activated_constraints_set(btx)

        btx.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(btx) == 0

        fin_fixed_vars = fixed_variables_set(btx)
        fin_act_consts = activated_constraints_set(btx)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, btx):
        assert (pytest.approx(5, abs=1e-3) ==
                value(btx.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(365, abs=1e-2) ==
                value(btx.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(151325, abs=1e2) ==
                value(btx.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(0, abs=1e-6) ==
                value(btx.fs.unit.work_mechanical[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, btx):
        assert abs(value(btx.fs.unit.inlet.flow_mol[0] -
                         btx.fs.unit.outlet.flow_mol[0])) <= 1e-6

        assert abs(btx.fs.unit.outlet.flow_mol[0] *
                   (btx.fs.unit.control_volume.properties_in[0]
                    .enth_mol_phase['Liq'] -
                    btx.fs.unit.control_volume.properties_out[0]
                    .enth_mol_phase['Liq'])) <= 1e-6

    @pytest.mark.ui
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
                "compressor": True})

        return m

    @pytest.mark.build
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet.vars) == 3
        assert hasattr(iapws.fs.unit.inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet, "pressure")

        assert hasattr(iapws.fs.unit, "outlet")
        assert len(iapws.fs.unit.outlet.vars) == 3
        assert hasattr(iapws.fs.unit.outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet, "pressure")

        assert hasattr(iapws.fs.unit, "work_mechanical")
        assert hasattr(iapws.fs.unit, "deltaP")
        assert isinstance(iapws.fs.unit.ratioP, Var)
        assert isinstance(iapws.fs.unit.ratioP_calculation, Constraint)

        assert isinstance(iapws.fs.unit.efficiency_isentropic, Var)
        assert isinstance(iapws.fs.unit.work_isentropic, Var)

        assert hasattr(iapws.fs.unit, "properties_isentropic")
        assert isinstance(iapws.fs.unit.isentropic_pressure, Constraint)
        assert isinstance(iapws.fs.unit.isentropic_material, Constraint)
        assert isinstance(iapws.fs.unit.isentropic, Constraint)
        assert isinstance(iapws.fs.unit.isentropic_energy_balance, Constraint)
        assert isinstance(iapws.fs.unit.actual_work, Constraint)

        assert number_variables(iapws) == 14
        assert number_total_constraints(iapws) == 9
        assert number_unused_variables(iapws) == 0

    def test_dof(self, iapws):
        iapws.fs.unit.inlet.flow_mol[0].fix(100)
        iapws.fs.unit.inlet.enth_mol[0].fix(4000)
        iapws.fs.unit.inlet.pressure[0].fix(101325)

        iapws.fs.unit.deltaP.fix(50000)
        iapws.fs.unit.efficiency_isentropic.fix(0.9)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, iapws):
        orig_fixed_vars = fixed_variables_set(iapws)
        orig_act_consts = activated_constraints_set(iapws)

        iapws.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(iapws) == 0

        fin_fixed_vars = fixed_variables_set(iapws)
        fin_act_consts = activated_constraints_set(iapws)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, iapws):
        # Check that outlet and isentropic pressure are equal
        assert pytest.approx(
            value(iapws.fs.unit.properties_isentropic[0].pressure), 1e-6) == \
            value(iapws.fs.unit.outlet.pressure[0])

        # Check that inlet and isentropic entropies are equal
        assert pytest.approx(
            value(iapws.fs.unit.properties_isentropic[0].entr_mol), 1e-6) == \
            value(iapws.fs.unit.control_volume.properties_in[0].entr_mol)

        assert pytest.approx(100, abs=1e-5) == \
            value(iapws.fs.unit.outlet.flow_mol[0])

        assert pytest.approx(4002, abs=1e0) == \
            value(iapws.fs.unit.outlet.enth_mol[0])

        assert pytest.approx(151325, abs=1e2) == \
            value(iapws.fs.unit.outlet.pressure[0])

        assert pytest.approx(101.43796915073504, abs=1e-1) == \
            value(iapws.fs.unit.work_mechanical[0])

        assert pytest.approx(91.29417223566153, abs=1e-1) == \
            value(iapws.fs.unit.work_isentropic[0])

        # For verification, check outlet and isentropic temperatures
        assert pytest.approx(326.170, 1e-5) == \
            value(iapws.fs.unit.control_volume.properties_out[0].temperature)

        assert pytest.approx(326.170, 1e-5) == \
            value(iapws.fs.unit.properties_isentropic[0].temperature)

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, iapws):
        assert abs(value(iapws.fs.unit.inlet.flow_mol[0] -
                         iapws.fs.unit.outlet.flow_mol[0])) <= 1e-6

        assert abs(value(
                iapws.fs.unit.outlet.flow_mol[0] *
                (iapws.fs.unit.inlet.enth_mol[0] -
                 iapws.fs.unit.outlet.enth_mol[0]) +
                iapws.fs.unit.work_mechanical[0])) <= 1e-6

    @pytest.mark.ui
    def test_report(self, iapws):
        iapws.fs.unit.report()


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = PressureChanger(default={
                "property_package": m.fs.properties,
                "thermodynamic_assumption": ThermodynamicAssumption.pump,
                "compressor": False})

        return m

    @pytest.mark.build
    def test_build(self, sapon):
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert hasattr(sapon.fs.unit, "work_mechanical")
        assert hasattr(sapon.fs.unit, "deltaP")
        assert isinstance(sapon.fs.unit.ratioP, Var)
        assert isinstance(sapon.fs.unit.ratioP_calculation, Constraint)

        assert isinstance(sapon.fs.unit.efficiency_pump, Var)
        assert isinstance(sapon.fs.unit.work_fluid, Var)
        assert isinstance(sapon.fs.unit.fluid_work_calculation, Constraint)
        assert isinstance(sapon.fs.unit.actual_work, Constraint)

        assert number_variables(sapon) == 21
        assert number_total_constraints(sapon) == 11
        assert number_unused_variables(sapon) == 0

    def test_dof(self, sapon):
        sapon.fs.unit.inlet.flow_vol[0].fix(1e-3)
        sapon.fs.unit.inlet.temperature[0].fix(320)
        sapon.fs.unit.inlet.pressure[0].fix(101325)
        sapon.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.deltaP.fix(-20000)
        sapon.fs.unit.efficiency_pump.fix(0.9)

        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, sapon):
        orig_fixed_vars = fixed_variables_set(sapon)
        orig_act_consts = activated_constraints_set(sapon)

        sapon.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(sapon) == 0

        fin_fixed_vars = fixed_variables_set(sapon)
        fin_act_consts = activated_constraints_set(sapon)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, sapon):
        assert pytest.approx(1e-3, abs=1e-6) == \
            value(sapon.fs.unit.outlet.flow_vol[0])

        assert pytest.approx(55388.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "H2O"])
        assert pytest.approx(100.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "NaOH"])
        assert pytest.approx(100.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"])
        assert pytest.approx(0.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"])
        assert pytest.approx(0.0, abs=1e-2) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "Ethanol"])

        assert pytest.approx(320.0, abs=1e-1) == \
            value(sapon.fs.unit.outlet.temperature[0])

        assert pytest.approx(81325, abs=1e2) == \
            value(sapon.fs.unit.outlet.pressure[0])

        assert pytest.approx(-18.0, abs=1e-2) == \
            value(sapon.fs.unit.work_mechanical[0])
        assert pytest.approx(-20.0, abs=1e-2) == \
            value(sapon.fs.unit.work_fluid[0])

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, sapon):
        assert abs(value(
                sapon.fs.unit.outlet.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.inlet.temperature[0] -
                 sapon.fs.unit.outlet.temperature[0]) +
                sapon.fs.unit.work_mechanical[0])) <= 1e-4

    @pytest.mark.ui
    def test_report(self, sapon):
        sapon.fs.unit.report()



from .process_base import ProcessBlockData, useDefault, MaterialFlowBasis
from .process_block import ProcessBlock, declare_process_block_class
from .unit_model import UnitModelBlockData, UnitModelBlock
from .flowsheet_model import FlowsheetBlockData, FlowsheetBlock
from .property_base import (StateBlockData, PhysicalParameterBlock,
                            StateBlock)
from .reaction_base import (ReactionBlockDataBase, ReactionParameterBlock,
                            ReactionBlockBase)
from .control_volume_base import (ControlVolumeBlockData, CONFIG_Template,
                                  MaterialBalanceType, EnergyBalanceType,
                                  MomentumBalanceType, FlowDirection)
from .control_volume0d import ControlVolume0DBlock
from .control_volume1d import ControlVolume1DBlock



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
General Cubic Equation of State property package with VLE calucations.
Cubic formulation and pure component property correlations from:
"The Properties of Gases and Liquids, 4th Edition", Reid, Prausnitz and Poling,
McGraw-Hill, 1987

Smooth Vapor-Liquid Equilibrium formulation from:
"A Smooth, Square Flash Formulation for Equation-Oriented Flowsheet
Optimization", Burgard et al., Proceedings of the 13 the International
Symposium on Process Systems Engineering – PSE 2018, July 1-5, 2018, San Diego

All results have been cross-referenced against other sources.
"""

# Chages the divide behavior to not do integer division
from __future__ import division

# Import Python libraries
import math
import os
from enum import Enum

# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           exp,
                           Expression,
                           ExternalFunction,
                           log,
                           NonNegativeReals,
                           Set,
                           SolverFactory,
                           sqrt,
                           Param,
                           PositiveReals,
                           value,
                           Var)
from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.core.util.initialization import (solve_indexed_blocks,
                                            fix_state_vars,
                                            revert_state_vars)
from idaes.core.util.exceptions import BurntToast
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_activated_equalities)
from idaes import lib_directory
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition


# Set up logger
_log = getIdaesLogger(__name__)


# Set path to root finder .so file
_so = os.path.join(lib_directory, "cubic_roots.so")


def cubic_roots_available():
    """Make sure the compiled cubic root functions are available. Yes, in
    Windows the .so extention is still used.
    """
    return os.path.isfile(_so)


class CubicEoS(Enum):
    PR = 0
    SRK = 1


EoS_param = {
        CubicEoS.PR: {'u': 2, 'w': -1, 'omegaA': 0.45724, 'coeff_b': 0.07780},
        CubicEoS.SRK: {'u': 1, 'w': 0, 'omegaA': 0.42748, 'coeff_b': 0.08664}
        }


@declare_process_block_class("CubicParameterBlock")
class CubicParameterData(PhysicalParameterBlock):
    """
    General Property Parameter Block Class
    """
    # Config block for the _IdealStateBlock
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare("valid_phase", ConfigValue(
        default=('Vap', 'Liq'),
        domain=In(['Liq', 'Vap', ('Vap', 'Liq'), ('Liq', 'Vap')]),
        description="Flag indicating the valid phase",
        doc="""Flag indicating the valid phase for a given set of
conditions, and thus corresponding constraints  should be included,
**default** - ('Vap', 'Liq').
**Valid values:** {
**'Liq'** - Liquid only,
**'Vap'** - Vapor only,
**('Vap', 'Liq')** - Vapor-liquid equilibrium,
**('Liq', 'Vap')** - Vapor-liquid equilibrium,}"""))

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(CubicParameterData, self).build()

        self.state_block_class = CubicStateBlock

        # List of valid phases in property package
        if self.config.valid_phase == ('Liq', 'Vap') or \
                self.config.valid_phase == ('Vap', 'Liq'):
            self.phase_list = Set(initialize=['Liq', 'Vap'],
                                  ordered=True)
        elif self.config.valid_phase == 'Liq':
            self.phase_list = Set(initialize=['Liq'])
        else:
            self.phase_list = Set(initialize=['Vap'])

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {'flow_mol': {'method': None, 'units': 'mol/s'},
             'mole_frac_comp': {'method': None, 'units': 'none'},
             'temperature': {'method': None, 'units': 'K'},
             'pressure': {'method': None, 'units': 'Pa'},
             'flow_mol_phase': {'method': None, 'units': 'mol/s'},
             'dens_mol_phase': {'method': '_dens_mol_phase',
                                'units': 'mol/m^3'},
             'dens_mass_phase': {'method': '_dens_mass_phase',
                                 'units': 'kg/m^3'},
             'pressure_sat': {'method': '_pressure_sat', 'units': 'Pa'},
             'mole_frac_phase_comp': {'method': '_mole_frac_phase',
                                      'units': 'no unit'},
             'enth_mol_phase': {'method': '_enth_mol_phase',
                                'units': 'J/mol'},
             'enth_mol': {'method': '_enth_mol', 'units': 'J/mol'},
             'entr_mol_phase': {'method': '_entr_mol_phase',
                                'units': 'J/mol'},
             'entr_mol': {'method': '_entr_mol', 'units': 'J/mol.K'},
             'fug_phase_comp': {'method': '_fug_phase', 'units': 'Pa'},
             'fug_coeff_phase_comp': {'method': '_fug_coeff_phase',
                                      'units': '-'},
             'gibbs_mol_phase': {'method': '_gibbs_mol_phase',
                                 'units': 'J/mol'},
             'mw': {'method': '_mw', 'units': 'kg/mol'},
             'mw_phase': {'method': '_mw_phase', 'units': 'kg/mol'},
             'temperature_bubble': {'method': '_temperature_bubble',
                                    'units': 'K'},
             'temperature_dew': {'method': '_temperature_dew',
                                 'units': 'K'},
             'pressure_bubble': {'method': '_pressure_bubble',
                                 'units': 'Pa'},
             'pressure_dew': {'method': '_pressure_dew',
                              'units': 'Pa'}})

        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})


class _CubicStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, state_args=None, state_vars_fixed=False,
                   hold_state=False, outlvl=5,
                   solver='ipopt', optarg={'tol': 1e-8}):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess. Expected keys in state_args dict are:
                         * flow_mol
                         * mole_frac_comp (dict with components as keys)
                         * pressure
                         * temperature
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default=None)
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = getInitLogger(blk.name, outlvl)

        init_log.info(5, 'Starting initialization')

        # Deactivate the constraints specific for outlet block i.e.
        # when defined state is False
        for k in blk.keys():
            if blk[k].config.defined_state is False:
                blk[k].sum_mole_frac_out.deactivate()

        # Fix state variables if not already fixed
        if state_vars_fixed is False:
            flags = fix_state_vars(blk, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in blk.keys():
                if degrees_of_freedom(blk[k]) != 0:
                    raise Exception("State vars fixed but degrees of freedom "
                                    "for state block is not zero during "
                                    "initialization.")
        # Set solver options
        if optarg is None:
            sopt = {'tol': 1e-8}
        else:
            sopt = optarg

        opt = SolverFactory('ipopt')
        opt.options = sopt

        # ---------------------------------------------------------------------
        # If present, initialize bubble and dew point calculations
        # Antoine equation
        def antoine_P(b, j, T):
            return 1e5*10**(b._params.antoine[j, '1'] -
                            b._params.antoine[j, '2'] /
                            (T + b._params.antoine[j, '3']))

        # Bubble temperature initialization
        for k in blk.keys():
            if hasattr(blk[k], "_mole_frac_tbub"):
                Tbub0 = 0
                for j in blk[k]._params.component_list:
                    Tbub0 += value(
                            blk[k].mole_frac_comp[j] *
                            (blk[k]._params.antoine[j, '2'] /
                             (blk[k]._params.antoine[j, '1'] -
                              math.log10(value(blk[k].pressure*1e-5))) -
                             blk[k]._params.antoine[j, '3']))

                err = 1
                counter = 0

                while err > 1e-2 and counter < 100:
                    f = value(sum(antoine_P(blk[k], j, Tbub0) *
                                  blk[k].mole_frac_comp[j]
                                  for j in blk[k]._params.component_list) -
                              blk[k].pressure)
                    df = value(sum(
                            blk[k].mole_frac_comp[j] *
                            blk[k]._params.antoine[j, '2'] *
                            math.log(10)*antoine_P(blk[k], j, Tbub0) /
                            (Tbub0 + blk[k]._params.antoine[j, '3'])**2
                            for j in blk[k]._params.component_list))

                    if f/df > 20:
                        Tbub1 = Tbub0 - 20
                    elif f/df < -20:
                        Tbub1 = Tbub0 + 20
                    else:
                        Tbub1 = Tbub0 - f/df

                    err = abs(Tbub1 - Tbub0)
                    Tbub0 = Tbub1
                    counter += 1

                blk[k].temperature_bubble.value = Tbub0

                for j in blk[k]._params.component_list:
                    blk[k]._mole_frac_tbub[j].value = value(
                            blk[k].mole_frac_comp[j]*blk[k].pressure /
                            antoine_P(blk[k], j, Tbub0))

        # Dew temperature initialization
        for k in blk.keys():
            if hasattr(blk[k], "_mole_frac_tdew"):
                Tdew0 = 0
                for j in blk[k]._params.component_list:
                    Tdew0 += value(
                            blk[k].mole_frac_comp[j] *
                            (blk[k]._params.antoine[j, '2'] /
                             (blk[k]._params.antoine[j, '1'] -
                              math.log10(value(blk[k].pressure*1e-5))) -
                             blk[k]._params.antoine[j, '3']))

                err = 1
                counter = 0

                while err > 1e-2 and counter < 100:
                    f = value(blk[k].pressure *
                              sum(blk[k].mole_frac_comp[j] /
                                  antoine_P(blk[k], j, Tdew0)
                                  for j in blk[k]._params.component_list) - 1)
                    df = -value(blk[k].pressure*math.log(10) *
                                sum(blk[k].mole_frac_comp[j] *
                                    blk[k]._params.antoine[j, '2'] /
                                    ((Tdew0 +
                                      blk[k]._params.antoine[j, '3'])**2 *
                                    antoine_P(blk[k], j, Tdew0))
                                    for j in blk[k]._params.component_list))

                    if f/df > 20:
                        Tdew1 = Tdew0 - 20
                    elif f/df < -20:
                        Tdew1 = Tdew0 + 20
                    else:
                        Tdew1 = Tdew0 - f/df

                    err = abs(Tdew1 - Tdew0)
                    Tdew0 = Tdew1
                    counter += 1

                blk[k].temperature_dew.value = Tdew0

                for j in blk[k]._params.component_list:
                    blk[k]._mole_frac_tdew[j].value = value(
                            blk[k].mole_frac_comp[j]*blk[k].pressure /
                            antoine_P(blk[k], j, Tdew0))

        # Bubble pressure initialization
        for k in blk.keys():
            if hasattr(blk[k], "_mole_frac_pbub"):
                blk[k].pressure_bubble.value = value(
                        sum(blk[k].mole_frac_comp[j] *
                            antoine_P(blk[k], j, blk[k].temperature)
                            for j in blk[k]._params.component_list))

                for j in blk[k]._params.component_list:
                    blk[k]._mole_frac_pbub[j].value = value(
                            blk[k].mole_frac_comp[j] *
                            antoine_P(blk[k], j, blk[k].temperature) /
                            blk[k].pressure_bubble)

                blk[k].pressure_bubble.display()
                blk[k]._mole_frac_pbub.display()

        # Dew pressure initialization
        for k in blk.keys():
            if hasattr(blk[k], "_mole_frac_pdew"):
                blk[k].pressure_dew.value = value(
                        sum(1/(blk[k].mole_frac_comp[j] /
                               antoine_P(blk[k], j, blk[k].temperature))
                            for j in blk[k]._params.component_list))

                for j in blk[k]._params.component_list:
                    blk[k]._mole_frac_pdew[j].value = value(
                            blk[k].mole_frac_comp[j]*blk[k].pressure_bubble /
                            antoine_P(blk[k], j, blk[k].temperature))

        # Solve bubble and dew point constraints
        cons_count = 0
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                # Deactivate all property constraints
                if c.local_name not in ("eq_pressure_dew",
                                        "eq_pressure_bubble",
                                        "eq_temperature_dew",
                                        "eq_temperature_bubble",
                                        "_sum_mole_frac_tbub",
                                        "_sum_mole_frac_tdew",
                                        "_sum_mole_frac_pbub",
                                        "_sum_mole_frac_pdew"):
                    c.deactivate()
            cons_count += number_activated_equalities(blk[k])

        if cons_count > 0:
            results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        else:
            results = None
        init_log.log(4,
                     "Dew and bubble point init: {}."
                     .format(condition(results)))

        # ---------------------------------------------------------------------
        # If flash, initialize T1 and Teq
        for k in blk.keys():
            if ((blk[k]._params.config.valid_phase == ('Liq', 'Vap')) or
                    (blk[k]._params.config.valid_phase == ('Vap', 'Liq'))):
                blk[k]._t1.value = max(blk[k].temperature.value,
                                       blk[k].temperature_bubble.value)
                blk[k]._teq.value = min(blk[k]._t1.value,
                                        blk[k].temperature_dew.value)
        init_log.log(4, "Equilibrium temperature init complete.")

        # ---------------------------------------------------------------------
        # Initialize flow rates and compositions
        # TODO : This will need to be generalised more when we move to a
        # modular implementation
        for k in blk.keys():
            if blk[k]._params.config.valid_phase == "Liq":
                blk[k].flow_mol_phase['Liq'].value = \
                    blk[k].flow_mol.value

                for j in blk[k]._params.component_list:
                    blk[k].mole_frac_phase_comp['Liq', j].value = \
                        blk[k].mole_frac_comp[j].value

            elif blk[k]._params.config.valid_phase == "Vap":
                blk[k].flow_mol_phase['Vap'].value = \
                    blk[k].flow_mol.value

                for j in blk[k]._params.component_list:
                    blk[k].mole_frac_phase_comp['Vap', j].value = \
                        blk[k].mole_frac_comp[j].value

            else:
                if blk[k].temperature.value > blk[k].temperature_dew.value:
                    # Pure vapour
                    blk[k].flow_mol_phase["Vap"].value = blk[k].flow_mol.value
                    blk[k].flow_mol_phase["Vap"].value = \
                        1e-5*blk[k].flow_mol.value

                    for j in blk[k]._params.component_list:
                        blk[k].mole_frac_phase_comp['Vap', j].value = \
                            blk[k].mole_frac_comp[j].value
                        blk[k].mole_frac_phase_comp['Liq', j].value = \
                            blk[k]._mole_frac_tdew[j].value
                elif blk[k].temperature.value < \
                        blk[k].temperature_bubble.value:
                    # Pure liquid
                    blk[k].flow_mol_phase["Vap"].value = \
                        1e-5*blk[k].flow_mol.value
                    blk[k].flow_mol_phase["Vap"].value = blk[k].flow_mol.value

                    for j in blk[k]._params.component_list:
                        blk[k].mole_frac_phase_comp['Vap', j].value = \
                            blk[k]._mole_frac_tbub[j].value
                        blk[k].mole_frac_phase_comp['Liq', j].value = \
                            blk[k].mole_frac_comp[j].value
                else:
                    # Two-phase
                    # TODO : Try to find some better guesses than this
                    blk[k].flow_mol_phase["Vap"].value = \
                        0.5*blk[k].flow_mol.value
                    blk[k].flow_mol_phase["Vap"].value = \
                        0.5*blk[k].flow_mol.value

                    for j in blk[k]._params.component_list:
                        blk[k].mole_frac_phase_comp['Vap', j].value = \
                            blk[k].mole_frac_comp[j].value
                        blk[k].mole_frac_phase_comp['Liq', j].value = \
                            blk[k].mole_frac_comp[j].value

        # ---------------------------------------------------------------------
        # Solve phase equilibrium constraints
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                # Activate equilibrium constraints
                if c.local_name in ("total_flow_balance",
                                    "component_flow_balances",
                                    "equilibrium_constraint",
                                    "sum_mole_frac",
                                    "_t1_constraint",
                                    "_teq_constraint"):
                    c.activate()

        results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        init_log.log(4,
                     "Phase equilibrium init: {}.".format(condition(results)))

        # ---------------------------------------------------------------------
        # Initialize other properties
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                # Activate all constraints except sum_mole_frac_out
                if c.local_name not in ("sum_mole_frac_out"):
                    c.activate()

        results = solve_indexed_blocks(opt, [blk], tee=init_tee(init_log))
        init_log.log(4, "Property init: {}.".format(condition(results)))

        # ---------------------------------------------------------------------
        # Return state to initial conditions
        for k in blk.keys():
            if (blk[k].config.defined_state is False):
                blk[k].sum_mole_frac_out.activate()

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

        init_log.log(5, "Initialization complete.")

    def release_state(blk, flags, outlvl=6):
        '''
        Method to relase state variables fixed during initialization.
        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        '''
        if flags is None:
            return

        init_log = getInitLogger(blk.name, outlvl)

        # Unfix state variables
        revert_state_vars(blk, flags)

        init_log.info(5, '{} states released.'.format(blk.name))

@declare_process_block_class("CubicStateBlock",
                             block_class=_CubicStateBlock)
class CubicStateBlockData(StateBlockData):
    """An general property package for cubic equations of state with VLE."""

    def build(self):
        """Callable method for Block construction."""
        super(CubicStateBlockData, self).build()

        # Add state variables
        self.flow_mol = Var(initialize=1.0,
                            domain=NonNegativeReals,
                            doc='Component molar flowrate [mol/s]')
        self.mole_frac_comp = Var(
                self._params.component_list,
                bounds=(0, None),
                initialize=1/len(self._params.component_list),
                doc='Mixture mole fractions [-]')
        self.pressure = Var(initialize=101325,
                            domain=NonNegativeReals,
                            doc='State pressure [Pa]')
        self.temperature = Var(initialize=298.15,
                               domain=NonNegativeReals,
                               doc='State temperature [K]')

        # Add supporting variables
        self.flow_mol_phase = Var(self._params.phase_list,
                                  initialize=0.5,
                                  domain=NonNegativeReals,
                                  doc='Phase molar flow rates [mol/s]')

        self.mole_frac_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            initialize=1/len(self._params.component_list),
            bounds=(0, None),
            doc='Phase mole fractions [-]')

        if self._params.config.valid_phase == "Liq":
            self._make_liq_phase_eq()
        elif self._params.config.valid_phase == "Vap":
            self._make_vap_phase_eq()
        elif ((self._params.config.valid_phase == ('Liq', 'Vap')) or
                (self._params.config.valid_phase == ('Vap', 'Liq'))):
            self._make_flash_eq()
        else:
            raise BurntToast("{} found unexpected value for valid_phases. "
                             "Please contact the "
                             "IDAES developers with this bug."
                             .format(self.name))

    def _make_liq_phase_eq(self):
        # Add equilibrium temperature - in this case the state temperature
        self._teq = Expression(expr=self.temperature)

        # Add supporting equations for Cubic EoS
        self.common_cubic()

        def rule_total_mass_balance(b):
            return b.flow_mol_phase['Liq'] == b.flow_mol
        self.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.mole_frac_comp[i] == b.mole_frac_phase_comp['Liq', i]
        self.component_flow_balances = Constraint(self._params.component_list,
                                                  rule=rule_comp_mass_balance)

        if self.config.defined_state is False:
            # applied at outlet only
            self.sum_mole_frac_out = Constraint(
                expr=1 == sum(self.mole_frac_comp[i]
                              for i in self._params.component_list))

    def _make_vap_phase_eq(self):
        # Add equilibrium temperature - in this case the state temperature
        self._teq = Expression(expr=self.temperature)

        # Add supporting equations for Cubic EoS
        self.common_cubic()

        def rule_total_mass_balance(b):
            return b.flow_mol_phase['Vap'] == b.flow_mol
        self.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.mole_frac_comp[i] == b.mole_frac_phase_comp['Vap', i]
        self.component_flow_balances = Constraint(self._params.component_list,
                                                  rule=rule_comp_mass_balance)

        if self.config.defined_state is False:
            # applied at outlet only
            self.sum_mole_frac_out = \
                Constraint(expr=1 == sum(self.mole_frac_comp[i]
                           for i in self._params.component_list))

    def _make_flash_eq(self):
        """
        Implementation of smooth VLE formulation.
        See module header for reference.
        """
        def rule_total_mass_balance(b):
            return b.flow_mol_phase['Liq'] + \
                b.flow_mol_phase['Vap'] == b.flow_mol
        self.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.flow_mol*b.mole_frac_comp[i] == \
                b.flow_mol_phase['Liq']*b.mole_frac_phase_comp['Liq', i] + \
                b.flow_mol_phase['Vap']*b.mole_frac_phase_comp['Vap', i]
        self.component_flow_balances = Constraint(self._params.component_list,
                                                  rule=rule_comp_mass_balance)

        def rule_mole_frac(b):
            return sum(b.mole_frac_phase_comp['Liq', i]
                       for i in b._params.component_list) -\
                sum(b.mole_frac_phase_comp['Vap', i]
                    for i in b._params.component_list) == 0
        self.sum_mole_frac = Constraint(rule=rule_mole_frac)

        if self.config.defined_state is False:
            # applied at outlet only
            self.sum_mole_frac_out = \
                Constraint(expr=1 == sum(self.mole_frac_comp[i]
                           for i in self._params.component_list))

        # Definition of equilibrium temperature for smooth VLE
        self._teq = Var(initialize=self.temperature.value,
                        doc='Temperature for calculating '
                            'phase equilibrium')
        self._t1 = Var(initialize=self.temperature.value,
                       doc='Intermediate temperature for calculating Teq')

        self.eps_1 = Param(default=0.01,
                           mutable=True,
                           doc='Smoothing parameter for Teq')
        self.eps_2 = Param(default=0.0005,
                           mutable=True,
                           doc='Smoothing parameter for Teq')

        # Add supporting equations for Cubic EoS
        self.common_cubic()

        # PSE paper Eqn 13
        def rule_t1(b):
            return b._t1 == 0.5 * \
                (b.temperature + b.temperature_bubble +
                 sqrt((b.temperature - b.temperature_bubble)**2 +
                      b.eps_1**2))
        self._t1_constraint = Constraint(rule=rule_t1)

        # PSE paper Eqn 14
        # TODO : Add option for supercritical extension
        def rule_teq(b):
            return b._teq == 0.5 * (b._t1 + b.temperature_dew -
                                    sqrt((b._t1 - b.temperature_dew)**2 +
                                         b.eps_2**2))
        self._teq_constraint = Constraint(rule=rule_teq)

        def rule_tr_eq(b, i):
            return b._teq / b._params.temperature_crit[i]
        self._tr_eq = Expression(self._params.component_list,
                                 rule=rule_tr_eq,
                                 doc='Component reduced temperatures [-]')

        def rule_equilibrium(b, i):
            return (b._log_equilibrium_cubic("Vap", i) ==
                    b._log_equilibrium_cubic("Liq", i))
        self.equilibrium_constraint = \
            Constraint(self._params.component_list, rule=rule_equilibrium)

# -----------------------------------------------------------------------------
# Property Methods
    def _dens_mol_phase(self):
        self.dens_mol_phase = Var(self._params.phase_list,
                                  doc="Molar density [mol/m^3]")

        def rule_dens_mol_phase(b, p):
            if p == 'Vap':
                return b._dens_mol_vap()
            else:
                return b._dens_mol_liq()
        self.eq_dens_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_dens_mol_phase)

    def _dens_mass_phase(self):
        self.dens_mass_phase = Var(self._params.phase_list,
                                   doc="Mass density [kg/m^3]")

        def rule_dens_mass_phase(b, p):
            if p == 'Vap':
                return b._dens_mass_vap()
            else:
                return b._dens_mass_liq()
        self.eq_dens_mass_phase = Constraint(self._params.phase_list,
                                             rule=rule_dens_mass_phase)

    def _enth_mol_phase(self):
        self.enth_mol_phase = Var(
            self._params.phase_list,
            doc='Phase molar specific enthalpies [J/mol]')

        def rule_enth_mol_phase(b, p):
            if p == "Vap":
                return b.enth_mol_phase[p] == b._enth_mol_vap()
            else:
                return b.enth_mol_phase[p] == b._enth_mol_liq()
        self.eq_enth_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_enth_mol_phase)

    def _enth_mol(self):
        self.enth_mol = Var(
            doc='Mixture molar specific enthalpies [J/mol]')

        def rule_enth_mol(b):
            return b.enth_mol*b.flow_mol == sum(
                    b.flow_mol_phase[p]*b.enth_mol_phase[p]
                    for p in b._params.phase_list)
        self.eq_enth_mol = Constraint(rule=rule_enth_mol)

    def _entr_mol(self):
        self.entr_mol = Var(
            doc='Mixture molar specific entropies [J/mol.K]')

        def rule_entr_mol(b):
            return b.entr_mol*b.flow_mol == sum(
                    b.flow_mol_phase[p]*b.entr_mol_phase[p]
                    for p in b._params.phase_list)
        self.eq_entr_mol = Constraint(rule=rule_entr_mol)

    def _entr_mol_phase(self):
        self.entr_mol_phase = Var(
            self._params.phase_list,
            doc='Phase molar specific entropies [J/mol.K]')

        def rule_entr_mol_phase(b, p):
            if p == "Vap":
                return b.entr_mol_phase[p] == b._entr_mol_vap()
            else:
                return b.entr_mol_phase[p] == b._entr_mol_liq()
        self.eq_entr_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_entr_mol_phase)

    def _fug_phase(self):
        def rule_fug_phase(b, p, j):
            if p == 'Vap':
                return b._fug_vap(j)
            else:
                return b._fug_liq(j)
        self.fug_phase_comp = Expression(self._params.phase_list,
                                         self._params.component_list,
                                         rule=rule_fug_phase)

    def _fug_coeff_phase(self):
        def rule_fug_coeff_phase(b, p, j):
            if p == 'Vap':
                return b._fug_coeff_vap(j)
            else:
                return b._fug_coeff_liq(j)
        self.fug_coeff_phase_comp = Expression(self._params.phase_list,
                                               self._params.component_list,
                                               rule=rule_fug_coeff_phase)

    def _gibbs_mol_phase(self):
        self.gibbs_mol_phase = Var(
            self._params.phase_list,
            doc='Phase molar specific Gibbs energy [J/mol]')

        def rule_gibbs_mol_phase(b, p):
            return b.gibbs_mol_phase[p] == (
                    b.enth_mol_phase[p] - b.temperature*b.entr_mol_phase[p])
        self.eq_gibbs_mol_phase = Constraint(self._params.phase_list,
                                             rule=rule_gibbs_mol_phase)

    def _mw(self):
        def rule_mw(b):
            return sum(b.mw_phase[p] for p in b._params.phase_list)
        self.mw = Expression(rule=rule_mw)

    def _mw_phase(self):
        def rule_mw_phase(b, p):
            return sum(b.mole_frac_phase_comp[p, j]*b._params.mw_comp[j]
                       for j in b._params.component_list)
        self.mw_phase = Expression(self._params.phase_list,
                                   rule=rule_mw_phase)

# -----------------------------------------------------------------------------
# General Methods
    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        if j in self._params.component_list:
            return self.flow_mol_phase[p] * self.mole_frac_phase_comp[p, j]
        else:
            return 0

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        return self.flow_mol_phase[p] * self.enth_mol_phase[p]

    def get_material_density_terms(self, p, j):
        """Create material density terms."""
        if j in self._params.component_list:
            return self.dens_mol_phase[p] * self.mole_frac_phase_comp[p, j]
        else:
            return 0

    def get_energy_density_terms(self, p):
        """Create energy density terms."""
        return self.dens_mol_phase[p] * self.enth_mol_phase[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        """Define state vars."""
        return {"flow_mol": self.flow_mol,
                "mole_frac_comp": self.mole_frac_comp,
                "temperature": self.temperature,
                "pressure": self.pressure}

    def define_display_vars(b):
        return {"Molar Flowrate": b.flow_mol,
                "Mole Fractions": b.cmole_frac_comp,
                "Temperature": b.temperature,
                "Pressure": b.pressure}

    def model_check(blk):
        """Model checks for property block."""
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error('{} Temperature set below lower bound.'
                       .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error('{} Temperature set above upper bound.'
                       .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error('{} Pressure set above upper bound.'.format(blk.name))

# -----------------------------------------------------------------------------
# Bubble and Dew Points
    def _temperature_bubble(self):
        self.temperature_bubble = Var(
                doc="Bubble point temperature (K)")

        self._mole_frac_tbub = Var(
                self._params.component_list,
                initialize=1/len(self._params.component_list),
                bounds=(0, None),
                doc="Vapor mole fractions at bubble point")

        self._sum_mole_frac_tbub = Constraint(
                expr=1e3 == 1e3*sum(self._mole_frac_tbub[j]
                                    for j in self._params.component_list))

        def rule_bubble_temp(b, j):
            return log(b.mole_frac_comp[j]) + log(b.bubble_temp_liq(j)) == \
                    log(b._mole_frac_tbub[j]) + log(b.bubble_temp_vap(j))
        self.eq_temperature_bubble = Constraint(self._params.component_list,
                                                rule=rule_bubble_temp)

    def _temperature_dew(self):
        self.temperature_dew = Var(
                doc="Dew point temperature (K)")

        self._mole_frac_tdew = Var(
                self._params.component_list,
                initialize=1/len(self._params.component_list),
                bounds=(0, None),
                doc="Liquid mole fractions at dew point")

        self._sum_mole_frac_tdew = Constraint(
                expr=1e3 == 1e3*sum(self._mole_frac_tdew[j]
                                    for j in self._params.component_list))

        def rule_dew_temp(b, j):
            return log(b._mole_frac_tdew[j]) + log(b.dew_temp_liq(j)) == \
                    log(b.mole_frac_comp[j]) + log(b.dew_temp_vap(j))
        self.eq_temperature_dew = Constraint(self._params.component_list,
                                             rule=rule_dew_temp)

    def _pressure_bubble(self):
        self.pressure_bubble = Var(
                domain=PositiveReals,
                doc="Bubble point pressure (Pa)")

        self._mole_frac_pbub = Var(
                self._params.component_list,
                initialize=1/len(self._params.component_list),
                bounds=(0, None),
                doc="Vapor mole fractions at bubble point")

        self._sum_mole_frac_pbub = Constraint(
                expr=1e3 == 1e3*sum(self._mole_frac_pbub[j]
                                    for j in self._params.component_list))

        def rule_bubble_pres(b, j):
            return log(b.mole_frac_comp[j]) + log(b.bubble_pres_liq(j)) == \
                    log(b._mole_frac_pbub[j]) + log(b.bubble_pres_vap(j))
        self.eq_pressure_bubble = Constraint(self._params.component_list,
                                             rule=rule_bubble_pres)

    def _pressure_dew(self):
        self.pressure_dew = Var(
                domain=PositiveReals,
                doc="Dew point pressure (Pa)")

        self._mole_frac_pdew = Var(
                self._params.component_list,
                initialize=1/len(self._params.component_list),
                bounds=(0, None),
                doc="Liquid mole fractions at dew point")

        self._sum_mole_frac_pdew = Constraint(
                expr=1e3 == 1e3*sum(self._mole_frac_pdew[j]
                                    for j in self._params.component_list))

        def rule_dew_press(b, j):
            return log(b._mole_frac_pdew[j]) + log(b.dew_press_liq(j)) == \
                    log(b.mole_frac_comp[j]) + log(b.dew_press_vap(j))
        self.eq_pressure_dew = Constraint(self._params.component_list,
                                          rule=rule_dew_press)

# -----------------------------------------------------------------------------
# Liquid phase properties
    def _vol_mol_liq(b):
        return b._vol_mol_cubic("Liq")

    def _dens_mol_liq(b):
        return b._dens_mol_cubic("Liq")

    def _dens_mass_liq(b):
        return b._dens_mass_cubic("Liq")

    def _fug_liq(b, j):
        return b._fug_cubic("Liq", j)

    def _fug_coeff_liq(b, j):
        return b._fug_coeff_cubic("Liq", j)

    def _enth_mol_liq(b):
        return b._enth_mol_cubic("Liq")

    def _enth_mol_liq_ig(b):
        return b._enth_mol_ig("Liq")

    def _entr_mol_liq(b):
        return b._entr_mol_cubic("Liq")

    def _entr_mol_liq_ig(b):
        return b._entr_mol_ig("Liq")

    def bubble_temp_liq(b, j):
        def a(k):
            return (b.omegaA*((b._params.gas_const *
                               b._params.temperature_crit[k])**2 /
                              b._params.pressure_crit[k]) *
                    ((1+b.fw[k]*(1-sqrt(b.temperature_bubble /
                                 b._params.temperature_crit[k])))**2))

        am = sum(sum(b.mole_frac_comp[i]*b.mole_frac_comp[j] *
                     sqrt(a(i)*a(j))*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b.mole_frac_comp[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure/(b._params.gas_const*b.temperature_bubble)**2
        B = bm*b.pressure/(b._params.gas_const*b.temperature_bubble)

        delta = (2*sqrt(a(j))/am *
                 sum(b.mole_frac_comp[i]*sqrt(a(i))*(1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_liq(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

    def dew_temp_liq(b, j):
        def a(k):
            return (b.omegaA*((b._params.gas_const *
                               b._params.temperature_crit[k])**2 /
                              b._params.pressure_crit[k]) *
                    ((1+b.fw[k]*(1-sqrt(b.temperature_dew /
                                 b._params.temperature_crit[k])))**2))

        am = sum(sum(b._mole_frac_tdew[i]*b._mole_frac_tdew[j] *
                     sqrt(a(i)*a(j))*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b._mole_frac_tdew[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure/(b._params.gas_const*b.temperature_dew)**2
        B = bm*b.pressure/(b._params.gas_const*b.temperature_dew)

        delta = (2*sqrt(a(j))/am *
                 sum(b._mole_frac_tdew[i]*sqrt(a(i))*(1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_liq(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

    def bubble_pres_liq(b, j):
        am = sum(sum(b.mole_frac_comp[i]*b.mole_frac_comp[j] *
                     sqrt(b.a[i]*b.a[j])*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b.mole_frac_comp[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure_bubble/(b._params.gas_const*b.temperature)**2
        B = bm*b.pressure_bubble/(b._params.gas_const*b.temperature)

        delta = (2*sqrt(b.a[j])/am *
                 sum(b.mole_frac_comp[i]*sqrt(b.a[i])*(1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_liq(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

    def dew_press_liq(b, j):
        am = sum(sum(b._mole_frac_pdew[i]*b._mole_frac_pdew[j] *
                     sqrt(b.a[i]*b.a[j])*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b._mole_frac_pdew[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure_dew/(b._params.gas_const*b.temperature)**2
        B = bm*b.pressure_dew/(b._params.gas_const*b.temperature)

        delta = (2*sqrt(b.a[j])/am *
                 sum(b._mole_frac_pdew[i]*sqrt(b.a[i]) *
                     (1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_liq(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

# -----------------------------------------------------------------------------
# Vapour phase properties
    def _vol_mol_vap(b):
        return b._vol_mol_cubic("Vap")

    def _dens_mol_vap(b):
        return b._dens_mol_cubic("Vap")

    def _dens_mass_vap(b):
        return b._dens_mass_cubic("Vap")

    def _fug_vap(b, j):
        return b._fug_cubic("Vap", j)

    def _fug_coeff_vap(b, j):
        return b._fug_coeff_cubic("Vap", j)

    def _enth_mol_vap(b):
        return b._enth_mol_cubic("Vap")

    def _enth_mol_vap_ig(b):
        return b._enth_mol_ig("Vap")

    def _entr_mol_vap(b):
        return b._entr_mol_cubic("Vap")

    def _entr_mol_vap_ig(b):
        return b._entr_mol_ig("Vap")

    def bubble_temp_vap(b, j):
        def a(k):
            return (b.omegaA*((b._params.gas_const *
                               b._params.temperature_crit[k])**2 /
                              b._params.pressure_crit[k]) *
                    ((1+b.fw[k]*(1-sqrt(b.temperature_bubble /
                                 b._params.temperature_crit[k])))**2))

        am = sum(sum(b._mole_frac_tbub[i]*b._mole_frac_tbub[j] *
                     sqrt(a(i)*a(j))*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b._mole_frac_tbub[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure/(b._params.gas_const*b.temperature_bubble)**2
        B = bm*b.pressure/(b._params.gas_const*b.temperature_bubble)

        delta = (2*sqrt(a(j))/am *
                 sum(b._mole_frac_tbub[i]*sqrt(a(i))*(1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_vap(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

    def dew_temp_vap(b, j):
        def a(k):
            return (b.omegaA*((b._params.gas_const *
                               b._params.temperature_crit[k])**2 /
                              b._params.pressure_crit[k]) *
                    ((1+b.fw[k]*(1-sqrt(b.temperature_dew /
                                 b._params.temperature_crit[k])))**2))

        am = sum(sum(b.mole_frac_comp[i]*b.mole_frac_comp[j] *
                     sqrt(a(i)*a(j))*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b.mole_frac_comp[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure/(b._params.gas_const*b.temperature_dew)**2
        B = bm*b.pressure/(b._params.gas_const*b.temperature_dew)

        delta = (2*sqrt(a(j))/am *
                 sum(b.mole_frac_comp[i]*sqrt(a(i))*(1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_vap(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

    def bubble_pres_vap(b, j):
        am = sum(sum(b._mole_frac_pbub[i]*b._mole_frac_pbub[j] *
                     sqrt(b.a[i]*b.a[j])*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b._mole_frac_pbub[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure_bubble/(b._params.gas_const*b.temperature)**2
        B = bm*b.pressure_bubble/(b._params.gas_const*b.temperature)

        delta = (2*sqrt(b.a[j])/am *
                 sum(b._mole_frac_pbub[i]*sqrt(b.a[i]) *
                     (1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_vap(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

    def dew_press_vap(b, j):
        am = sum(sum(b.mole_frac_comp[i]*b.mole_frac_comp[j] *
                     sqrt(b.a[i]*b.a[j])*(1-b._params.kappa[i, j])
                     for j in b._params.component_list)
                 for i in b._params.component_list)
        bm = sum(b.mole_frac_comp[i]*b.b[i] for i in b._params.component_list)

        A = am*b.pressure_dew/(b._params.gas_const*b.temperature)**2
        B = bm*b.pressure_dew/(b._params.gas_const*b.temperature)

        delta = (2*sqrt(b.a[j])/am *
                 sum(b.mole_frac_comp[i]*sqrt(b.a[i])*(1-b._params.kappa[j, i])
                     for i in b._params.component_list))

        Z = b.proc_Z_vap(b._ext_func_param, A, B)

        return exp((b.b[j]/bm*(Z-1)*(B*b.EoS_p) -
                   log(Z-B)*(B*b.EoS_p) +
                   A*(b.b[j]/bm - delta) *
                   log((2*Z + B*(b.EoS_u + b.EoS_p)) /
                       (2*Z + B*(b.EoS_u-b.EoS_p))))/(B*b.EoS_p))

# -----------------------------------------------------------------------------
# Common Cubic Functions
# All of these equations drawn from Properties of Gases and Liquids
# Quantities appended with _eq represent calcuations at equilibrium temperature
    def common_cubic(blk):
        if hasattr(blk, "omegaA"):
            return

        blk.omegaA = EoS_param[blk._params.cubic_type]['omegaA']

        blk.EoS_Bc = EoS_param[blk._params.cubic_type]['coeff_b']
        blk.EoS_u = EoS_param[blk._params.cubic_type]['u']
        blk.EoS_w = EoS_param[blk._params.cubic_type]['w']
        blk.EoS_p = sqrt(blk.EoS_u**2 - 4*blk.EoS_w)

        # Create expressions for coefficients
        def func_fw(b, j):
            if b._params.cubic_type == CubicEoS.PR:
                return 0.37464 + 1.54226*b._params.omega[j] - \
                       0.26992*b._params.omega[j]**2
            elif b._params.cubic_type == CubicEoS.SRK:
                return 0.48 + 1.574*b._params.omega[j] - \
                       0.176*b._params.omega[j]**2
            else:
                raise BurntToast(
                        "{} received unrecognised cubic type. This should "
                        "never happen, so please contact the IDAES developers "
                        "with this bug.".format(b.name))
        blk.fw = Param(blk._params.component_list,
                       initialize=func_fw,
                       doc='EoS S factor')

        def func_b(b, j):
            return b.EoS_Bc*b._params.gas_const *\
                   b._params.temperature_crit[j]/b._params.pressure_crit[j]
        blk.b = Param(blk._params.component_list,
                      initialize=func_b,
                      doc='Component b coefficient')

        def func_a(b, j):
            return (b.omegaA*((b._params.gas_const *
                               b._params.temperature_crit[j])**2 /
                              b._params.pressure_crit[j]) *
                    ((1+b.fw[j]*(1-sqrt(b.temperature /
                                        b._params.temperature_crit[j])))**2))
        blk.a = Expression(blk._params.component_list,
                           rule=func_a,
                           doc='Component a coefficient')

        def func_a_eq(b, j):
            return (b.omegaA*((b._params.gas_const *
                               b._params.temperature_crit[j])**2 /
                              b._params.pressure_crit[j]) *
                    ((1+b.fw[j]*(1-sqrt(b._teq /
                                        b._params.temperature_crit[j])))**2))
        blk._a_eq = Expression(blk._params.component_list,
                               rule=func_a_eq,
                               doc='Component a coefficient at Teq')

        def rule_am(b, p):
            return sum(sum(
                b.mole_frac_phase_comp[p, i]*b.mole_frac_phase_comp[p, j] *
                sqrt(b.a[i]*b.a[j])*(1-b._params.kappa[i, j])
                for j in b._params.component_list)
                for i in b._params.component_list)
        blk.am = Expression(blk._params.phase_list, rule=rule_am)

        def rule_am_eq(b, p):
            return sum(sum(
                b.mole_frac_phase_comp[p, i]*b.mole_frac_phase_comp[p, j] *
                sqrt(b._a_eq[i]*b._a_eq[j])*(1-b._params.kappa[i, j])
                for j in b._params.component_list)
                for i in b._params.component_list)
        blk._am_eq = Expression(blk._params.phase_list, rule=rule_am_eq)

        def rule_bm(b, p):
            return sum(b.mole_frac_phase_comp[p, i]*b.b[i]
                       for i in b._params.component_list)
        blk.bm = Expression(blk._params.phase_list, rule=rule_bm)

        def rule_A(b, p):
            return (b.am[p]*b.pressure /
                    (b._params.gas_const*b.temperature)**2)
        blk.A = Expression(blk._params.phase_list, rule=rule_A)

        def rule_B(b, p):
            return (b.bm[p]*b.pressure /
                    (b._params.gas_const*b.temperature))
        blk.B = Expression(blk._params.phase_list, rule=rule_B)

        def rule_A_eq(b, p):
            return (b._am_eq[p]*b.pressure /
                    (b._params.gas_const*b._teq)**2)
        blk._A_eq = Expression(blk._params.phase_list, rule=rule_A_eq)

        def rule_B_eq(b, p):
            return (b.bm[p]*b.pressure /
                    (b._params.gas_const*b._teq))
        blk._B_eq = Expression(blk._params.phase_list, rule=rule_B_eq)

        blk.proc_Z_liq = ExternalFunction(library=_so,
                                          function="ceos_z_liq")
        blk.proc_Z_vap = ExternalFunction(library=_so,
                                          function="ceos_z_vap")
        blk.proc_Z_liq_x = ExternalFunction(library=_so,
                                            function="ceos_z_liq_extend")
        blk.proc_Z_vap_x = ExternalFunction(library=_so,
                                            function="ceos_z_vap_extend")

        def rule_delta(b, p, i):
            # See pg. 145 in Properties of Gases and Liquids
            return (2*sqrt(blk.a[i])/b.am[p] *
                    sum(b.mole_frac_phase_comp[p, j]*sqrt(blk.a[j]) *
                        (1-b._params.kappa[i, j])
                        for j in b._params.component_list))
        blk.delta = Expression(blk._params.phase_list,
                               blk._params.component_list,
                               rule=rule_delta)

        def rule_delta_eq(b, p, i):
            # See pg. 145 in Properties of Gases and Liquids
            return (2*sqrt(blk._a_eq[i])/b._am_eq[p] *
                    sum(b.mole_frac_phase_comp[p, j]*sqrt(blk._a_eq[j]) *
                        (1-b._params.kappa[i, j])
                        for j in b._params.component_list))
        blk._delta_eq = Expression(blk._params.phase_list,
                                   blk._params.component_list,
                                   rule=rule_delta_eq)

        def rule_dadT(b, p):
            # See pg. 102 in Properties of Gases and Liquids
            return -((b._params.gas_const/2)*sqrt(b.omegaA) *
                     sum(sum(b.mole_frac_phase_comp[p, i] *
                             b.mole_frac_phase_comp[p, j] *
                             (1-b._params.kappa[i, j]) *
                             (b.fw[j]*sqrt(b.a[i] *
                                           b._params.temperature_crit[j] /
                                           b._params.pressure_crit[j]) +
                              b.fw[i]*sqrt(b.a[j] *
                                           b._params.temperature_crit[i] /
                                           b._params.pressure_crit[i]))
                             for j in b._params.component_list)
                         for i in b._params.component_list) /
                     sqrt(b.temperature))
        blk.dadT = Expression(blk._params.phase_list, rule=rule_dadT)

        blk._ext_func_param = Param(default=blk._params.cubic_type.value)

        def rule_compress_fact(b, p):
            if p == "Vap":
                return b.proc_Z_vap(b._ext_func_param, b.A[p], b.B[p])
            else:
                return b.proc_Z_liq(b._ext_func_param, b.A[p], b.B[p])
        blk.compress_fact_phase = Expression(blk._params.phase_list,
                                             rule=rule_compress_fact)

        def rule_compress_fact_eq(b, p):
            if p == "Vap":
                return b.proc_Z_vap(b._ext_func_param, b._A_eq[p], b._B_eq[p])
            else:
                return b.proc_Z_liq(b._ext_func_param, b._A_eq[p], b._B_eq[p])
        blk._compress_fact_eq = Expression(blk._params.phase_list,
                                           rule=rule_compress_fact_eq)

    def _vol_mol_cubic(b, p):
        return (b.pressure*b.vol_mol_phase[p] ==
                b.compress_fact_phase[p]*b._params.gas_const*b.temperature)

    def _dens_mol_cubic(b, p):
        return b.pressure == (b.dens_mol_phase[p]*b.compress_fact_phase[p] *
                              b._params.gas_const*b.temperature)

    def _dens_mass_cubic(b, p):
        return b.dens_mass_phase[p] == b.dens_mol_phase[p]*b.mw_phase[p]

    def _fug_cubic(b, p, j):
        return b.mole_frac_phase_comp[p, j]*b.pressure * \
               b.fug_coeff_phase_comp[p, j]

    def _fug_coeff_cubic(b, p, j):
        # See pg. 145 in Properties of Gases and Liquids
        return exp((b.b[j]/b.bm[p]*(b.compress_fact_phase[p]-1) *
                    (b.B[p]*b.EoS_p) -
                    log(b.compress_fact_phase[p]-b.B[p]) *
                    (b.B[p]*b.EoS_p) +
                    b.A[p]*(b.b[j]/b.bm[p] - b.delta[p, j]) *
                    log((2*b.compress_fact_phase[p] +
                         b.B[p]*(b.EoS_u + b.EoS_p)) /
                        (2*b.compress_fact_phase[p] +
                         b.B[p]*(b.EoS_u-b.EoS_p)))) /
                   (b.B[p]*b.EoS_p))

    def _log_equilibrium_cubic(b, p, j):
        # See pg. 145 in Properties of Gases and Liquids
        return ((b.b[j]/b.bm[p]*(b._compress_fact_eq[p]-1) *
                 (b._B_eq[p]*b.EoS_p) -
                 log(b._compress_fact_eq[p]-b._B_eq[p]) *
                 (b._B_eq[p]*b.EoS_p) +
                 b._A_eq[p]*(b.b[j]/b.bm[p] - b._delta_eq[p, j]) *
                 log((2*b._compress_fact_eq[p] +
                      b._B_eq[p]*(b.EoS_u + b.EoS_p)) /
                     (2*b._compress_fact_eq[p] +
                      b._B_eq[p]*(b.EoS_u-b.EoS_p)))) /
                (b._B_eq[p]*b.EoS_p) + log(b.mole_frac_phase_comp[p, j]))

    def _enth_mol_cubic(b, p):
        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        return (((b.temperature*b.dadT[p] - b.am[p]) *
                 log((2*b.compress_fact_phase[p] + b.B[p]*(b.EoS_u+b.EoS_p)) /
                     (2*b.compress_fact_phase[p] + b.B[p]*(b.EoS_u-b.EoS_p))) +
                 b._params.gas_const*b.temperature *
                 (b.compress_fact_phase[p]-1)*b.bm[p]*b.EoS_p) /
                (b.bm[p]*b.EoS_p) + b._enth_mol_ig(p))

    def _enth_mol_ig(b, p):
        return sum(b.mole_frac_phase_comp[p, j] *
                   (b._enth_mol_comp_ig(j) + b._params.enth_mol_form_ref[j])
                   for j in b._params.component_list)

    def _entr_mol_cubic(b, p):
        # See pg. 102 in Properties of Gases and Liquids
        return ((b._params.gas_const*log(
                    (b.compress_fact_phase[p]-b.B[p]) /
                    b.compress_fact_phase[p])*b.bm[p]*b.EoS_p +
                 b._params.gas_const*log(b.compress_fact_phase[p] *
                                         b._params.pressure_ref/b.pressure) *
                 b.bm[p]*b.EoS_p +
                 b.dadT[p]*log((2*b.compress_fact_phase[p] +
                                b.B[p]*(b.EoS_u + b.EoS_p)) /
                               (2*b.compress_fact_phase[p] +
                                b.B[p]*(b.EoS_u-b.EoS_p)))) /
                (b.bm[p]*b.EoS_p) + b._entr_mol_ig(p))

    def _entr_mol_ig(b, p):
        return sum(b.mole_frac_phase_comp[p, j] *
                   (b._entr_mol_comp_ig(j) + b._params.entr_mol_form_ref[j])
                   for j in b._params.component_list)

# -----------------------------------------------------------------------------
# Pure component properties
    def _enth_mol_comp_ig(b, j):
        return (
            (b._params.cp_ig[j, "4"]/4) *
            (b.temperature**4-b._params.temperature_ref**4) +
            (b._params.cp_ig[j, "3"]/3) *
            (b.temperature**3-b._params.temperature_ref**3) +
            (b._params.cp_ig[j, "2"]/2) *
            (b.temperature**2-b._params.temperature_ref**2) +
            b._params.cp_ig[j, "1"] *
            (b.temperature-b._params.temperature_ref))

    def _entr_mol_comp_ig(b, j):
        return ((b._params.cp_ig[j, '4']/3) *
                (b.temperature**3-b._params.temperature_ref**3) +
                (b._params.cp_ig[j, '3']/2) *
                (b.temperature**2-b._params.temperature_ref**2) +
                b._params.cp_ig[j, '2'] *
                (b.temperature-b._params.temperature_ref) +
                b._params.cp_ig[j, '1'] *
                log(b.temperature/b._params.temperature_ref))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import pytest

from idaes.core import FlowsheetBlock
from idaes.property_models.cubic_eos.cubic_prop_pack import \
    cubic_roots_available
from idaes.property_models.cubic_eos import BT_PR

from pyomo.environ import (ConcreteModel,
                           Objective,
                           SolverFactory,
                           TerminationCondition,
                           value)

from idaes.property_models.tests.test_harness import PropertyTestHarness
from idaes.core.util.testing import get_default_solver


# Set module level pyest marker
pytestmark = pytest.mark.cubic_root
prop_available = cubic_roots_available()


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
# Run test harness
class TestBasicLV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = BT_PR.BTParameterBlock
        self.param_args = {"valid_phase": ("Liq", "Vap")}
        self.prop_args = {}
        self.has_density_terms = True


class TestBasicL(PropertyTestHarness):
    def configure(self):
        self.prop_pack = BT_PR.BTParameterBlock
        self.param_args = {"valid_phase": "Liq"}
        self.prop_args = {}
        self.has_density_terms = True


class TestBasicV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = BT_PR.BTParameterBlock
        self.param_args = {"valid_phase": "Vap"}
        self.prop_args = {}
        self.has_density_terms = True


# -----------------------------------------------------------------------------
# Test robustness and some outputs
@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
class TestBTExample(object):
    def test_T_sweep(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.obj = Objective(expr=(m.fs.state.temperature - 510)**2)

        for logP in range(8, 13, 1):
            m.fs.obj.deactivate()

            m.fs.state.flow_mol.fix(100)
            m.fs.state.mole_frac_comp["benzene"].fix(0.5)
            m.fs.state.mole_frac_comp["toluene"].fix(0.5)
            m.fs.state.temperature.fix(300)
            m.fs.state.pressure.fix(10**(0.5*logP))

            m.fs.state.initialize(outlvl=0)

            m.fs.state.temperature.unfix()
            m.fs.obj.activate()

            solver = SolverFactory('ipopt')
            results = solver.solve(m, tee=True)

            assert results.solver.termination_condition == \
                TerminationCondition.optimal
            assert m.fs.state.flow_mol_phase["Liq"].value <= 1e-5

    def test_P_sweep(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        for T in range(370, 500, 25):
            m.fs.state.flow_mol.fix(100)
            m.fs.state.mole_frac_comp["benzene"].fix(0.5)
            m.fs.state.mole_frac_comp["toluene"].fix(0.5)
            m.fs.state.temperature.fix(T)
            m.fs.state.pressure.fix(1e5)

            m.fs.state.initialize(outlvl=0)

            solver = SolverFactory('ipopt')
            results = solver.solve(m)

            assert results.solver.termination_condition == \
                TerminationCondition.optimal

            while m.fs.state.pressure.value <= 1e6:
                m.fs.state.pressure.value = m.fs.state.pressure.value + 1e5
                solver = SolverFactory('ipopt')
                results = solver.solve(m)
                assert results.solver.termination_condition == \
                    TerminationCondition.optimal
                print(T, m.fs.state.pressure.value)

    def test_T350_P1_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(350)
        m.fs.state.pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state.enth_mol_phase
        m.fs.state.entr_mol_phase

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        solver.solve(m)

        assert pytest.approx(value(m.fs.state._teq), abs=1e-1) == 365
        assert 0.0035346 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Liq"]), 1e-5)
        assert 0.966749 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Vap"]), 1e-5)
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
                1e-5) == 0.894676
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
                1e-5) == 0.347566
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
                1e-5) == 0.971072
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
                1e-5) == 0.959791

        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
                1e-5) == 0.5
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
                1e-5) == 0.5
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
                1e-5) == 0.70584
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
                1e-5) == 0.29416

        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 38942.8
        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 78048.7
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -367.558
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -269.0553

    def test_T350_P5_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(350)
        m.fs.state.pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state.enth_mol_phase
        m.fs.state.entr_mol_phase

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        solver.solve(m)

        assert pytest.approx(value(m.fs.state._teq), 1e-5) == 431.47
        assert pytest.approx(
                value(m.fs.state.compress_fact_phase["Liq"]), 1e-5) == 0.01766
        assert pytest.approx(
                value(m.fs.state.compress_fact_phase["Vap"]), 1e-5) == 0.80245
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
                1e-5) == 0.181229
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
                1e-5) == 0.070601
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
                1e-5) == 0.856523
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
                1e-5) == 0.799237

        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
                1e-5) == 0.5
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
                1e-5) == 0.5
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
                1e-5) == 0.65415
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
                1e-5) == 0.34585

        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 38966.9
        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 75150.7
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -367.6064
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -287.3318

    def test_T450_P1_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(450)
        m.fs.state.pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state.enth_mol_phase
        m.fs.state.entr_mol_phase

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        solver.solve(m)

        assert pytest.approx(value(m.fs.state._teq), 1e-5) == 371.4
        assert 0.0033583 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Liq"]), 1e-5)
        assert 0.9821368 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Vap"]), 1e-5)
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
                1e-5) == 8.069323
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
                1e-5) == 4.304955
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
                1e-5) == 0.985365
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
                1e-5) == 0.979457

        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
                1e-5) == 0.29861
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
                1e-5) == 0.70139
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
                1e-5) == 0.5
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
                1e-5) == 0.5

        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 49441.2
        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 84175.1
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -333.836
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -247.385

    def test_T450_P5_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(450)
        m.fs.state.pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state.enth_mol_phase
        m.fs.state.entr_mol_phase

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        solver.solve(m)

        assert pytest.approx(value(m.fs.state._teq), 1e-5) == 436.93
        assert 0.0166181 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Liq"]), 1e-5)
        assert 0.9053766 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Vap"]), 1e-5)
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
                1e-5) == 1.63308
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
                1e-5) == 0.873213
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
                1e-5) == 0.927534
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
                1e-5) == 0.898324

        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
                1e-5) == 0.3488737
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
                1e-5) == 0.6511263
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
                1e-5) == 0.5
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
                1e-5) == 0.5

        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 51095.2
        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 83362.3
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -331.676
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -261.961

    def test_T368_P1_x5(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.5)
        m.fs.state.mole_frac_comp["toluene"].fix(0.5)
        m.fs.state.temperature.fix(368)
        m.fs.state.pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state.enth_mol_phase
        m.fs.state.entr_mol_phase

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        solver.solve(m)

        assert pytest.approx(value(m.fs.state._teq), 1e-5) == 368
        assert 0.003504 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Liq"]), 1e-5)
        assert 0.97 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Vap"]), 1e-5)
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
                1e-5) == 1.492049
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
                1e-5) == 0.621563
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
                1e-5) == 0.97469
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
                1e-5) == 0.964642

        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
                1e-5) == 0.4012128
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
                1e-5) == 0.5987872
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
                1e-5) == 0.6141738
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
                1e-5) == 0.3858262

        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 38235.1
        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 77155.4
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -364.856
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -267.892

    def test_T376_P1_x2(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={'dynamic': False})

        m.fs.props = BT_PR.BTParameterBlock(
                default={'valid_phase': ('Vap', 'Liq')})

        m.fs.state = m.fs.props.state_block_class(
                default={'parameters': m.fs.props,
                         'defined_state': True})

        m.fs.state.flow_mol.fix(100)
        m.fs.state.mole_frac_comp["benzene"].fix(0.2)
        m.fs.state.mole_frac_comp["toluene"].fix(0.8)
        m.fs.state.temperature.fix(376)
        m.fs.state.pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state.enth_mol_phase
        m.fs.state.entr_mol_phase

        m.fs.state.initialize(outlvl=0)

        solver = SolverFactory('ipopt')
        solver.solve(m)

        assert pytest.approx(value(m.fs.state._teq), 1e-5) == 376
        assert 0.00361333 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Liq"]), 1e-5)
        assert 0.968749 == pytest.approx(
                value(m.fs.state.compress_fact_phase["Vap"]), 1e-5)
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "benzene"]),
                1e-5) == 1.8394188
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Liq", "toluene"]),
                1e-5) == 0.7871415
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "benzene"]),
                1e-5) == 0.9763608
        assert pytest.approx(
                value(m.fs.state.fug_coeff_phase_comp["Vap", "toluene"]),
                1e-5) == 0.9663611

        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "benzene"]),
                1e-5) == 0.17342
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Liq", "toluene"]),
                1e-5) == 0.82658
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "benzene"]),
                1e-5) == 0.3267155
        assert pytest.approx(
                value(m.fs.state.mole_frac_phase_comp["Vap", "toluene"]),
                1e-5) == 0.6732845

        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Liq"]), 1e-5) == 31535.8
        assert pytest.approx(
                value(m.fs.state.enth_mol_phase["Vap"]), 1e-5) == 69175.3
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Liq"]), 1e-5) == -372.869
        assert pytest.approx(
                value(m.fs.state.entr_mol_phase["Vap"]), 1e-5) == -278.766



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import re


def almpywriter(data):
    # This funciton writes the file
    # - <almname>alm.py
    #    y=<fname>.f(X)
    # handle the multiple output
    if data['opts']['noutputs'] > 1:
        for output_name, mod_res in data['results']['model'].items():
            almpywriter_help(data, mod_res, output_name)
    else:
        almpywriter_help(data, data['results']['model'], data['labs']['savezlabels'][0])


def almpywriter_help(data, mod_res, output_name):
    # This funciton writes the file
    # - <almname>alm.py
    #    y=<fname>.f(X)
    # preliminary formating to get the model ready to write
    model = mod_res.split('=')[1]
    model = model + ' '
    tlist = ('sin', 'cos', 'log', 'exp')
    for tok in tlist:
        model = model.replace(tok, 'np.' + tok)
    model = model.replace('^', '**')

    with open(output_name + '.py', 'w') as r:
        r.write('import numpy as np\n')
        r.write('def f(*X):\n')
        i = 0
        for lab in data['labs']['savexlabels']:
            r.write('    ' + lab + '= X[' + str(i) + ']\n')
            i = i + 1
        r.write('    return ' + model + '\n')


def almcvwriter(data):
    # This funciton writes the file
    # - <almname>alm.py
    #    y=<fname>.f(X)
    # handle the multiple output
    if data['opts']['noutputs'] > 1:
        for output_name, mod_res in data['results']['model'].items():
            almcvwriter_help(data, mod_res, output_name)
    else:
        almcvwriter_help(data, data['results']['model'], data['labs']['savezlabels'][0])


def almcvwriter_help(data, mod_res, output_name):
    # This funciton writes the file
    # - <almname>cv.py
    #    y=<fname>.f(X,params)
    # print data['results']['model']
    model = mod_res.split('=')[1]
    model = model + ' '
    tlist = ('sin', 'cos', 'log', 'exp')
    for tok in tlist:
        model = model.replace(tok, 'np.' + tok)
    model = model.replace('^', '**')

    # remove numerical values of coefficients
    model = re.sub(r"(?:.\w)?([0-9]+\.[0-9]+)", 'b', model)
    model = re.sub(r"[E]{1}.\d{3}", '', model)
    model = model.replace('-', '+')

    tind = 0
    tstr = ''
    while 'b' in model:
        model = model.replace('b', 'B[' + str(tind) + ']', 1)
        tstr = tstr + ',B' + str(tind)
        tind = tind + 1
    with open(data['stropts']['almname'].split('.')[0] + 'cv.py', 'w') as r:
        line = 'def f(X, B):\n'
        r.write(line)
        r.write('    import numpy as np\n')
        i = 0
        for lab in data['labs']['savexlabels']:
            r.write('    ' + lab + '= X[' + str(i) + ']\n')
            i = i + 1
        r.write('    return ' + model + '\n')


def wrapwriter(sim):
    import os
    import inspect
    # This subroutine writes a temporary python file that is used
    # to wrap python-based function that lack ALAMO's input/output.txt I/O
    name = 'simwrapper'
    name = name + '.py'
    with open(name, 'w') as r:
        r.write("#!/usr/bin/python\n")
        r.write("def main():\n")
        r.write("    import " + inspect.getmodule(sim).__name__ + "\n")
        r.write("    infile = 'input.txt'\n")
        r.write("    outfile = 'output.txt'\n")
        r.write("    fin = open(infile, 'r')\n")
        r.write("    fout = open(outfile, 'w')\n")
        r.write("    newline = fin.readline()\n")
        r.write("    newlist = newline.split()\n")
        r.write("    n = int(newlist[0])\n")
        r.write("    for p in range(0, n):\n")
        r.write("        newline = fin.readline()\n")
        r.write("        newlist = newline.split()\n")
        r.write("        ninputs = len(newlist)\n")
        r.write("        x = [0] * (ninputs + 1)\n")
        r.write("        for k in range(0, ninputs):\n")
        r.write("            x[k] = float(newlist[k])\n")
        # r.write("        x[ninputs] = "+sim.__name__+"(x[:-1])\n")
        r.write("        x[ninputs] = " + inspect.getmodule(sim).__name__ 
                + '.' + sim.__name__ + "(*x[:-1])\n")
        # r.write("        x[ninputs] = "+str(sim)+"(x[:-1])\n")
        r.write("        for k in range(0, len(x)):\n")
        r.write("            fout.write(str(x[k]) + ' ')\n")
        r.write("        fout.write(' \\n')\n")
        r.write("\n\n")
        r.write("if __name__ == '__main__':\n")
        r.write("    main()\n")

    os.chmod(name, 509)
    return name



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.util module
"""
import datetime
import hashlib
import json
import logging
import os
import shutil
import sys
import time

import pytest

#
from idaes.dmf import util, resource
from .util import init_logging, TempDir

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


def test_strlist():
    input = [1, 2, 3]
    output = util.strlist(input, sep="/")
    assert output == "1/2/3"


def test_get_file():
    f1 = util.get_file(__file__)
    assert f1 is not None
    f2 = util.get_file(f1)
    assert f2 is not None
    assert f2.name == f1.name


def test_import_module():
    m = util.import_module("idaes.dmf.util")
    assert m is not None


def test_get_module_version():
    m = util.import_module("idaes.dmf.util")
    v1 = util.get_module_version(m)
    assert v1 is None
    m.__version__ = "foobar"
    try:
        util.get_module_version(m)
    except ValueError:
        pass
    else:
        assert False, "ValueError expected for {}".format(m.__version__)
    m.__version__ = "1.2.3-alpha"
    util.get_module_version(m)


def test_get_module_author():
    m = util.import_module("idaes.dmf.util")
    util.get_module_author(m)


def test_tempdir():
    with util.TempDir() as newdir:
        pass


def test_datetime_timestamp():
    ts = time.time()
    dt = datetime.datetime.fromtimestamp(ts)
    ts1 = util.datetime_timestamp(dt)
    assert pytest.approx(ts, ts1, 0.000001)


def test_mkdir_p():
    random_str = hashlib.sha1().hexdigest()
    # test absolute
    path = "/tmp/{}/idaes/dmf/util/mkdir_p".format(random_str)
    util.mkdir_p(path)
    assert os.path.exists(path)
    shutil.rmtree("/tmp/{}".format(random_str))
    # test relative
    path = "{}/idaes/dmf/util/mkdir_p".format(random_str)
    util.mkdir_p(path)
    assert os.path.exists(path)
    shutil.rmtree(random_str)


def test_is_jupyter_notebook():
    nbtext = """
    {
        "cells": [],
        "metadata": {},
        "nbformat": 4,
        "nbformat_minor": 2
    }
    """
    with TempDir() as d:
        f = open(os.path.join(d, "sample.txt"), "w")
        assert not util.is_jupyter_notebook(f.name)
        f = open(os.path.join(d, "null.ipynb"), "w")
        assert util.is_jupyter_notebook(f.name, check_contents=False)
        assert not util.is_jupyter_notebook(f.name)
        f = open(os.path.join(d, "basic.ipynb"), "w")
        f.write(nbtext)
        f.close()
        assert util.is_jupyter_notebook(f.name)
        f = open(os.path.join(d, "random.ipynb"), "w")
        f.write('{"random": [1,2,3]}')
        f.close()
        assert not util.is_jupyter_notebook(f.name)


def test_is_python():
    with TempDir() as d:
        f = open(os.path.join(d, "sample.txt"), "w")
        assert not util.is_python(f.name)
        f = open(os.path.join(d, "sample.py"), "w")
        assert util.is_python(f.name)


def test_is_resource_json():
    with TempDir() as d:
        f = open(os.path.join(d, "sample.txt"), "w")
        assert not util.is_resource_json(f.name)
        f = open(os.path.join(d, "sample.json"), "w")
        assert not util.is_resource_json(f.name)
        r = resource.Resource()
        json.dump(r.v, f)
        f.close()
        assert util.is_resource_json(f.name)
        assert not util.is_resource_json(f.name, max_bytes=1)
